
from MsgEvent import MsgEvent
import math
import random

MAX_PLANET_PLACEMENT_TRIES = 15

class Planet:
	def __init__(self, field, id, owner, indust, ships):
		self.field = field
		self.id = id
		self.seen = {}
		self.ships = { owner.id : (ships-indust, owner) }
		self.set_owner(owner, 0.0)
		self.indust = indust
		self.has_been_surprised = 0
		self.winch()

	def set_owner(self, owner, year):
		self.owner = owner
		self.seen[owner.id] = year, owner
		self.owner.bkeep_planet(self)
		for p in self.ships.keys():
			if self.ships[p][0] > 0:
				self.seen[p] = year, self.ships[p][1]

	def get_ships(self, pl):
		if not self.ships.has_key(pl.id):
			return 0
		else:
			return self.ships[pl.id][0]

	def set_ships(self, pl, sh):
		self.ships[pl.id] = sh, pl

	def get_location(self):
		return self.loc

	def get_selfasevent(self, evtype):
		return MsgEvent(evtype, planet=self, loc=self.loc, player=self.owner)

	def check_move(self, event):
		if self.loc == None or event.tg_planet.loc == None:
			event.player.send_event(
				MsgEvent('ERRS', name=event.evtype,
					 desc="Your planet's in the void.."))
			return 0
		if self.owner != event.player:
			event.player.send_event(
				MsgEvent('ERRP', name=event.evtype,
					 desc="You can only move your ships."))
			return 0
		if self.get_ships(event.player) < event.quant:
			event.player.send_event(
				MsgEvent('ERRP', name=event.evtype, 
					 desc="There aren't that many ships."))
			return 0
		if event.tg_planet == self:
			event.player.send_event(
				MsgEvent('ERRP', name=event.evtype, 
					 desc="That's not a move."))
			return 0
		if event.quant <= 0:
			event.player.send_event(
				MsgEvent('ERRC', name=event.evtype, 
					 desc="Bad ship count."))
			return 0
		return 1

	def record_move(self, event, eta):
		self.set_ships(event.player, self.get_ships(event.player) - event.quant)
		event.player.send_event(MsgEvent('MVOK', player=event.player, planet=self, \
						 tg_planet=event.tg_planet, change=eta, \
						 quant=self.get_ships(event.player)))

	def send_broadcast(self, event, pll=[]):
		for p in pll:
			p.send_event(event)
		for p in self.seen.values():
			if not p[1] in pll:
				p[1].send_event(event)

	def attackers(self, n):
		if n <= 1:
			return n
		p = 1.0 / (1.0 + math.log10(n))
		np = n * p
		m = random.gauss(np, math.sqrt(np*(1-p)))
		if m > n:
			return n
		elif m < 0:
			return 0
		else:
			return int(round(m))

	def surprise(self, event):
		spf = self.field.get_prob_factor(event)
		return (random.random() < (float(spf) / (event.quant + spf)))

	def battle(self, event, year, who):
		sh = [self.get_ships(self.owner), event.quant]
		pl = [self.owner, event.player]
		ev = MsgEvent('BATT', player=pl[1], planet=self, tg_player=pl[0], \
			      quant=sh[1], change=sh[0], which=who)
		self.send_broadcast(ev, pl)
		while sh[0] > 0 and sh[1] > 0:
			h = self.attackers(sh[who])
			if h >= sh[1-who]:
				h = sh[1-who]
			sh[1-who] = sh[1-who] - h
			ev = MsgEvent('ATTK', player=pl, planet=self, quant=sh, change=h, which=who)
			for p in pl:
				p.send_event(ev)
			who = 1 - who
		if sh[0] > 0:
			who = 0
		else:
			who = 1
		self.set_ships(self.owner, 0)
		self.set_owner(pl[who], year)
		self.set_ships(self.owner, sh[who])
		ev = MsgEvent('BWIN', player=pl[who], tg_player=pl[1-who], which=who, \
			      planet=self, quant=self.get_ships(self.owner))
		self.send_broadcast(ev, pl)
		pl[1-who].think_about_death()

	def recv_arrive(self, event, year):
		if not self.owner.is_enemy(event.player) and \
		   not event.player.is_attackable(self.owner):
			self.set_ships(event.player, self.get_ships(event.player) + event.quant)
			self.send_broadcast(MsgEvent('FORT', player=event.player, planet=self, \
						     quant=self.get_ships(event.player), change=event.quant))
		else:
			if self.has_been_surprised:
				surprise = 0
			else:
				surprise = self.surprise(event)
				self.has_been_surprised = surprise
			self.battle(event, year, surprise)
		self.send_broadcast(MsgEvent('EARR', player=self.owner, planet=self, \
					     quant=self.get_ships(event.player), tg_player=event.player))
		self.field.game.send_broadcast(MsgEvent('PLNT', planet=self, loc=self.loc, \
							player=self.owner))

	def yearly(self, year):
		for p in self.seen.keys():
			if self.seen[p][0] < year - self.field.spy_aging and self.ships[p][0] <= 0:
				self.seen[p][1].ubkeep_planet(self)
				del self.seen[p]
		self.set_owner(self.owner, year)
		self.has_been_surprised = 0
		self.set_ships(self.owner, self.get_ships(self.owner) + self.indust)

	def send_selfbroadcast(self, evtype):
		if evtype == 'PIND':
			ev = MsgEvent(evtype, planet=self, quant=self.indust)
		else:
			ev = MsgEvent(evtype, planet=self, player=self.owner, quant=self.get_ships(self.owner))
		self.send_broadcast(ev)

	def winch(self):
		def randto(max): return random.randint(0,max-1)
		tries = 0
		while tries < MAX_PLANET_PLACEMENT_TRIES:
			self.loc = tuple(map(randto,self.field.get_dimensions()))
			if not self.field.get_planetat(self.loc):
				break
			tries = tries + 1
		else:
			self.loc = None
		self.field.reg_planet(self)

